const { resolve } = require('path');
const webpack = require('webpack');
const { VueLoaderPlugin } = require('vue-loader');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const CopyWebpackPlugin = require('copy-webpack-plugin');
const ESLintPlugin = require('eslint-webpack-plugin');
// const BundleAnalyzerPlugin = require('webpack-bundle-analyzer').BundleAnalyzerPlugin;

const { ModuleFederationPlugin } = webpack.container;

const pkgInfo = require('./package.json');
const { name: appName, dependencies } = pkgInfo;

const port = (() => {
  return process.env.PORT || 8080;
})();
const devUrl = `http://localhost:${port}`;
const cwd = process.cwd();
const publicDir = resolve(cwd, 'public');
const isDev = process.env.NODE_ENV === 'development';

const getModuleRemotes = () => {
  let workforceRemote = 'workforce@http://heal.com/workforce/remoteEntry.js';
  if (process.env.WORKFORCE_APP) {
    workforceRemote = `workforce@http://${process.env.WORKFORCE_APP}/remoteEntry.js`;
  }

  let networkRemote = 'network@http://heal.com/network/remoteEntry.js';
  if (process.env.NETWORK_APP) {
    networkRemote = `network@http://${process.env.NETWORK_APP}/remoteEntry.js`;
  }
  return {
    workforce: workforceRemote,
    network: networkRemote,
  };
};

module.exports = {
  mode: 'development',
  devtool: false,
  entry: {
    main: resolve(cwd, 'src/main.js'),
  },
  output: {
    path: resolve(cwd, `dist/${appName}`),
    publicPath: isDev ? `${devUrl}/` : `http://heal.com/${appName}/`,
    filename: '[name].js', // 打包后输出文件的文件名
  },
  // 配置 webpack-dev-server 使用
  // https://www.webpackjs.com/configuration/dev-server/#devserverhot
  devServer: {
    static: {
      directory: publicDir, // 指定静态资源的路径
    },
    compress: true, // 是否压缩
    port, // 指定使用端口
    hot: true, // 是否热更新，那肯定啦...
    open: true, // 编译完成打开浏览器
    client: {
      logging: 'error', // 终端显示错误即可
      overlay: {
        // warnings 类型不要挡着页面，error才展示为浏览器页面
        errors: true,
        warnings: false,
      },
    },
    historyApiFallback: { disableDotRule: true, index: '/' },
    // proxy: { // 配置代理
    //   '/api': {
    //     target: 'http://localhost:3000',
    //     changeOrigin: true,
    //   },
    // },
  },
  stats: {
    preset: 'minimal',
  },
  optimization: {
    minimize: false,
    // 这样设置可以解决主应用开发时热更新失效（子应用不会），原因未知
    // 但是会影响到子应用开发时候的加载，因为对于模块内容的初始化被抽取出去到独自的 runtime.js
    // 但是子应用开发时候是没有加载这份单独出来的 runtime.js
    // 所以用标志位判断主应用开发模式再用 'single' 模式即可
    runtimeChunk: isDev ? 'single' : false,
    // splitChunks: {
    //   chunks: 'all',
    //   // minChunks: 3,
    //   // minSize: 0,
    //   cacheGroups: {
    //     default: {
    //       priority: 1,
    //       reuseExistingChunk: true,
    //       enforce: true
    //     }
    //   }
    // }
  },
  module: {
    rules: [
      {
        test: /\.js$/, // 解析 js 文件
        exclude: /node_modules/,
        use: 'babel-loader',
      },
      {
        test: /\.vue$/, // 解析 vue 文件
        exclude: /node_modules/,
        use: 'vue-loader', // 17版本会导致 css 打包报错
      },
      {
        test: /\.css$/, // 添加对样式表的处理
        use: [
          // 请只在生产环境下使用 CSS 提取，这将便于你在开发环境下进行热重载。
          isDev ? 'vue-style-loader' : MiniCssExtractPlugin.loader,
          'css-loader',
        ],
      },
      {
        test: /\.less$/, // 使用 less sass 这些要额外声明处理规则
        use: [
          isDev ? 'vue-style-loader' : MiniCssExtractPlugin.loader,
          'css-loader',
          'less-loader',
        ],
      },
      {
        test: /\.(png|jpg|gif)$/i,
        use: [
          {
            loader: 'url-loader',
            options: {
              // 新版本的file-loader选项esModule默认为true，需要将其设置为false，否则打包后src为[object Module]
              esModule: false,
              limit: 8192,
            },
          },
        ],
      },
    ],
  },
  resolve: {
    alias: {
      '@': resolve(cwd, 'src'),
    },
  },
  plugins: [
    // new BundleAnalyzerPlugin(),
    // 参考官方文档 https://vue-loader.vuejs.org/migrating.html#a-plugin-is-now-required
    // Vue-loader在15.*之后的版本都是 vue-loader的使用都是需要伴生 VueLoaderPlugin的,
    new VueLoaderPlugin(),
    // 编译时替换全局变量
    new webpack.DefinePlugin({
      BASE_URL: JSON.stringify('/'),
      // 'process.env.NODE_ENV': JSON.stringify(process.env.NODE_ENV),
    }),
    new ESLintPlugin({
      extensions: ['.js', '.vue'], // Default: 'js'
    }), // https://eslint.org/docs/latest/integrate/nodejs-api#-new-eslintoptions
    new MiniCssExtractPlugin({
      filename: 'style.css',
    }),
    // 用来将打包的 js 注入到 index.html
    new HtmlWebpackPlugin({
      inject: 'body',
      template: resolve(publicDir, 'index.html'),
    }),
    // 把 public 的文件拷贝过去 dist 文件
    new CopyWebpackPlugin({
      patterns: [
        {
          from: publicDir,
          globOptions: {
            ignore: [
              '**/index.html', // html 由 HtmlWebpackPlugin 生成，不需要拷贝
              '**/.DS_Store',
              '**/Thumbs.db',
            ],
          },
        },
      ],
    }),
    new ModuleFederationPlugin({
      name: appName,
      filename: 'remoteEntry.js',
      remotes: getModuleRemotes(),
      exposes: {
        './utils': './src/utils/index.js',
        './init': './src/bootstrap.js',
        './layout': './src/components/Layout.vue',
        './welcome': './src/components/Welcome.vue',
      },
      shared: {
        ...dependencies,
        vue: {
          eager: true,
          singleton: true,
          requiredVersion: dependencies.vue,
        },
        vuex: {
          eager: true,
          singleton: true,
          requiredVersion: dependencies.vuex,
        },
        'vue-router': {
          eager: true,
          singleton: true,
          requiredVersion: dependencies['vue-router'],
        },
      },
    }),
  ],
};
